@@ -352,9 +352,6 @@ module Stub = struct
     connection -> result_status -> result = "PQmakeEmptyPGresult_stub"
 
   external ntuples : result -> int = "PQntuples_stub" "noalloc"
-
-(* FIXME: switch to noalloc once PostgreSQL 8.2 is out for CentOS *)
-(*   external nparams : result -> int = "PQnparams_stub" "noalloc" *)
   external nparams : result -> int = "PQnparams_stub"
 
   external nfields : result -> int = "PQnfields_stub" "noalloc"
@@ -362,9 +359,6 @@ module Stub = struct
   external fnumber : result -> string -> int ="PQfnumber_stub" "noalloc"
   external fformat : result -> int -> FFormat.t = "PQfformat_stub" "noalloc"
   external ftype : result -> int -> oid = "PQftype_stub" "noalloc"
-
-(* FIXME: switch to noalloc once PostgreSQL 8.2 is out for CentOS *)
-(*   external paramtype : result -> int -> oid = "PQparamtype_stub" "noalloc" *)
   external paramtype : result -> int -> oid = "PQparamtype_stub"
 
   external fmod : result -> int -> int = "PQfmod_stub" "noalloc"
@@ -373,6 +367,9 @@ module Stub = struct
 
   external getvalue : result -> int -> int -> string = "PQgetvalue_stub"
 
+  external get_escaped_value :
+    result -> int -> int -> string = "PQgetescval_stub"
+
   external getisnull :
     result -> int -> int -> bool = "PQgetisnull_stub" "noalloc"
 
@@ -473,6 +470,7 @@ end
 (* Escaping *)
 
 external unescape_bytea_stub : string -> string = "PQunescapeBytea_stub"
+external unescape_bytea_9x_stub : string -> string = "PQunescapeBytea9x_stub"
 
 let unescape_bytea str =
   let str_len = String.length str in
@@ -481,58 +479,14 @@ let unescape_bytea str =
   else
     (* This is the new Postgresql 9.0 hex format for encoding bytea.  Older
        clients do not recognize this format, but servers send it by default. *)
-    let rec count_hex ~pos ~count =
-      if pos = str_len then count
-      else
-        match str.[pos] with
-        | ' ' | '\t' | '\n' | '\r' | '\b' -> count_hex ~pos:(pos + 1) ~count
-        | '0'..'9' | 'a'..'f' | 'A'..'F' ->
-            let pos = pos + 1 in
-            if pos = str_len then
-              failwith "Postgresql.unescape_bytea: incomplete hex code"
-            else
-              begin match str.[pos] with
-              | '0'..'9' | 'a'..'f' | 'A'..'F' ->
-                  count_hex ~pos:(pos + 1) ~count:(count + 1)
-              | _ ->
-                  failwith "Postgresql.unescape_bytea: short hex code"
-              end
-        | _ -> failwith "Postgresql.unescape_bytea: illegal hex code"
-    in
-    let n_grps = count_hex ~pos:2 ~count:0 in
-    let res = String.create n_grps in
-    let rec res_loop ~str_pos ~res_pos =
-      if res_pos = n_grps then res
-      else
-        match str.[str_pos] with
-        | ' ' | '\t' | '\n' | '\r' | '\b' ->
-            res_loop ~str_pos:(str_pos + 1) ~res_pos
-        | c ->
-            let unhexdigit = function
-              | '0'..'9' as c -> Char.code c - Char.code '0'
-              | 'a'..'f' as c -> Char.code c - Char.code 'a' + 10
-              | 'A'..'F' as c -> Char.code c - Char.code 'A' + 10
-              | _ -> assert false
-            in
-            let high = unhexdigit c lsl 4 in
-            let str_pos = str_pos + 1 in
-            let low = unhexdigit str.[str_pos] in
-            let res_char = Char.chr (high + low) in
-            res.[res_pos] <- res_char;
-            res_loop ~str_pos:(str_pos + 1) ~res_pos:(res_pos + 1)
-    in
-    res_loop ~str_pos:2 ~res_pos:0
-
+    unescape_bytea_9x_stub str
 
 (* Query results *)
 
 class result res =
   let nfields = Stub.nfields res in
   let ntuples = Stub.ntuples res in
-(* FIXME: get rid of laziness once PostgreSQL 8.2 is out for CentOS *)
-(*   let nparams = Stub.nparams res in *)
   let nparams = lazy (Stub.nparams res) in
-  let binary_tuples = Stub.binary_tuples res in
   let check_field field =
     if field < 0 || field >= nfields then
       raise (Error (Field_out_of_range (field, nfields))) in
@@ -549,7 +503,7 @@ object
   method ntuples = ntuples
   method nparams = Lazy.force nparams
   method nfields = nfields
-  method binary_tuples = binary_tuples
+  method binary_tuples = Stub.binary_tuples res
   method fname field = check_field field; Stub.fname res field
 
   method fnumber s =
@@ -576,6 +530,11 @@ object
     check_tuple tuple;
     Stub.getvalue res tuple field
 
+  method get_escaped_value tuple field =
+    check_field field;
+    check_tuple tuple;
+    Stub.get_escaped_value res tuple field
+
   method getisnull tuple field =
     check_field field; check_tuple tuple;
     Stub.getisnull res tuple field
